import os
from edge.image.image import ImRgba

from PIL import ImageDraw, ImageFont, Image
from edge.lane_detector import LaneDetector
from _args import RendererArgs, RaAnnoArgs, LrAnnoArgs
from _models import LrSceneItem, RaSceneItem, TrScene, TrSceneItem, FaceSceneItem
from _color_store import LrColorToutStore

from edge import eventlog
from .tl import ITlService

_FONT_PT_SIZE = 20

_RA_BOX_DIR = 1
_RA_BOX_SPEED = 2
_RA_BOX_LENGTH = 3
_RA_BOX_LANE = 4

_LR_BOX_DIR = 5
_LR_BOX_PLATE_TEXT = 6
_LR_BOX_LANE = 7

_LR_BOX_COORD_BBOX = 8

_RA_BOX_SPEED_WITHOUT_LENGTH = 8
_RA_BOX_DIR_WITHOUT_LENGTH = 9

_SYM_W = 12
_SYM_H = 17
_SYM_MARGIN_W = 6
_SYM_MARGIN_H = 3

_MODE_ALWAYS = 'always'
_MODE_NEVER = 'never'
_MODE_IF_PLATE = 'if-plate'

_LANE_UNDEFINED = 0


def _get_ra_board_box_ltrb(ref_point_xy, box_id):
    """
    :type ref_point_xy: (int, int)
    :type box_id: int
    :rtype: (int, int, int, int)
    """
    if box_id == _RA_BOX_DIR:
        ir = ref_point_xy[0] - 4 * _SYM_W - _SYM_MARGIN_W - _SYM_MARGIN_W
        il = ir - _SYM_MARGIN_W - 2 * _SYM_MARGIN_W

    elif box_id == _RA_BOX_SPEED:
        ir = ref_point_xy[0] - (_SYM_MARGIN_W / 2)
        il = ir - 4 * _SYM_W - _SYM_MARGIN_W / 2

    elif box_id == _RA_BOX_LENGTH:
        il = ref_point_xy[0] + (_SYM_MARGIN_W / 2)
        ir = il + 4 * _SYM_W + _SYM_MARGIN_W

    elif box_id == _RA_BOX_LANE:
        il = ref_point_xy[0] + 4 * _SYM_W + _SYM_MARGIN_W
        ir = il + 2 * _SYM_MARGIN_W + _SYM_W

    elif box_id == _RA_BOX_SPEED_WITHOUT_LENGTH:
        ir = ref_point_xy[0] + 2 * _SYM_W + _SYM_MARGIN_W
        il = ref_point_xy[0] - 2 * _SYM_W - _SYM_MARGIN_W / 4

    elif box_id == _RA_BOX_DIR_WITHOUT_LENGTH:
        ir = ref_point_xy[0] - 2 * _SYM_W - _SYM_MARGIN_W
        il = (ref_point_xy[0] + (_SYM_MARGIN_W / 2)) - 4 * _SYM_W - _SYM_MARGIN_W

    else:
        raise NotImplementedError

    it = ref_point_xy[1] + _SYM_MARGIN_H
    ib = it + _SYM_H + _SYM_MARGIN_H
    return il, it, ir, ib


def _get_lr_board_box_ltrb(ref_point_xy, box_id):
    """
    :type ref_point_xy: (int, int)
    :type box_id: int
    :rtype: (int, int, int, int)
    """
    if box_id == _LR_BOX_DIR:
        ir = ref_point_xy[0] - 4 * _SYM_W - _SYM_MARGIN_W - _SYM_MARGIN_W
        il = ir - _SYM_MARGIN_W - 2 * _SYM_MARGIN_W

    elif box_id == _LR_BOX_PLATE_TEXT:
        ir = ref_point_xy[0] + 4 * _SYM_W + _SYM_MARGIN_W
        il = ref_point_xy[0] - 4 * _SYM_W - _SYM_MARGIN_W

    elif box_id == _LR_BOX_LANE:
        il = ref_point_xy[0] + 4 * _SYM_W + _SYM_MARGIN_W + _SYM_MARGIN_W
        ir = il + _SYM_MARGIN_W + _SYM_W

    elif box_id == _LR_BOX_COORD_BBOX:
        ir = ref_point_xy[0] + 4 * _SYM_W + _SYM_MARGIN_W
        il = ref_point_xy[0] - 4 * _SYM_W - _SYM_MARGIN_W

    else:
        raise NotImplementedError

    if box_id == _LR_BOX_COORD_BBOX:
        it = ref_point_xy[1] - 2*_SYM_MARGIN_H - _SYM_H
        ib = it + _SYM_H + _SYM_MARGIN_H
    else:
        it = ref_point_xy[1] + _SYM_MARGIN_H
        ib = it + _SYM_H + _SYM_MARGIN_H
    return il, it, ir, ib


def _render_text(draw, bg_ltrb, fg_text, bg_color=(255, 255, 255), fg_color=(0, 0, 0)):
    """
    :type draw: PIL.ImageDraw.ImageDraw
    :type bg_ltrb: (int, int, int, int)
    :type fg_text: str | unicode
    :type bg_color: (int, int, int)
    :type fg_color: (int, int, int)
    """
    # render background
    draw.rectangle(bg_ltrb, fill=bg_color)

    # render foreground
    text_w, text_h = draw.getfont().getsize(fg_text)
    text_il = (bg_ltrb[0] + bg_ltrb[2]) / 2 - text_w / 2
    text_it = (bg_ltrb[1] + bg_ltrb[3]) / 2 - _FONT_PT_SIZE / 2

    draw.text((text_il, text_it), fg_text, fill=fg_color)


def _render_speed(draw, bg_ltrb, speed, yellow_speed, red_speed):
    """
    :type draw: PIL.ImageDraw.ImageDraw
    :type bg_ltrb: (int, int, int, int)
    :type speed: int
    :type yellow_speed: int
    :type red_speed: int
    """
    if red_speed and speed > red_speed:
        bg_color = (208, 100, 89)  # red

    elif yellow_speed and speed > yellow_speed:
        bg_color = (255, 214, 0)  # amber

    else:
        bg_color = (145, 186, 108)  # green

    _render_text(draw, bg_ltrb, fg_text=str(speed), bg_color=bg_color)


def _render_dir(draw, bg_ltrb, fg_dir):
    """
    :type draw: PIL.ImageDraw.ImageDraw
    :type bg_ltrb: (int, int, int, int)
    :type fg_dir: int
    """
    _DIR_APPROACHING = 11
    _DIR_RECEDING = 12

    if fg_dir == _DIR_APPROACHING:
        fg_text = u'\u2193'
        bg_color = (140, 200, 255)

    elif fg_dir == _DIR_RECEDING:
        fg_text = u'\u2191'
        bg_color = (215, 165, 255)

    else:
        fg_text = '?'
        bg_color = (255, 180, 90)

    _render_text(draw, bg_ltrb, fg_text, bg_color)


def _render_tl(draw, green_time, red_time, amber_time):
    """
    :type draw: ImageDraw.ImageDraw
    :type green_time: long | None
    :type red_time: long | None
    :type amber_time: long | None
    :rtype: None
    """
    red_ltrb = (720, 30, 728, 38)
    amber_ltrb = (720, 50, 728, 58)
    green_ltrb = (720, 70, 728, 78)

    if green_time is not None:
        draw.rectangle(green_ltrb, fill='green')

    if red_time is not None:
        draw.rectangle(red_ltrb, fill='red')

    if amber_time is not None:
        draw.rectangle(amber_ltrb, fill='yellow')


class Renderer:
    _FONT_ABS_PATH = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'ttf', 'Cousine-Bold.ttf')
    _FONT = ImageFont.truetype(_FONT_ABS_PATH, _FONT_PT_SIZE, encoding='utf-8')

    _PLATE_MARGIN_X = int(2 * _FONT_PT_SIZE / 10.0)
    _PLATE_MARGIN_Y = int(1 * _FONT_PT_SIZE / 10.0)
    _TEXT_BLACK_COLOR = (0, 0, 0)

    _RA_ITEM_RADIUS = 4

    _COLOR_BLACK = (0, 0, 0)
    _COLOR_WHITE = (255, 255, 255)
    _COLOR_YELLOW = (255, 255, 0)
    _COLOR_GREEN = (0, 255, 0)
    _COLOR_RED = (255, 0, 0)

    # http://www.stm.dp.ua/web-design/color-html.php
    _COLORS = [
        (255, 248, 220),  # Cornsilk
        (255, 228, 181),  # Moccasin
        (255, 245, 238),  # Seashell
        (240, 255, 240),  # Honeydew
        (230, 230, 250),  # lavender
        (240, 255, 255),  # Azure
        (255, 228, 225),  # MistyRose
        (152, 251, 152),  # PaleGreen
        (255, 215,   0),  # Gold
        (188, 143, 143),  # RosyBrown
        (245, 245, 220),  # Beige
        (245, 222, 179),  # Wheat
        (193, 255, 193),  # DarkSeaGreen1
        (84,  255, 159)   # SeaGreen1
    ]

    def __init__(self, args, color_store, lane_detector):
        """
        :type args: RendererArgs
        :type color_store: LrColorToutStore
        :type lane_detector: LaneDetector
        """
        self._color_store = color_store  # +
        self._lane_detector = lane_detector

        self._lr_anno_args = args.lr_anno
        self._ra_anno_args = args.ra_anno

    def render_tr_scene(self, tr_scene, im_rgba, tl_service):
        """
        :type tr_scene: TrScene
        :type im_rgba: ImRgba
        :type tl_service: ITlService
        :rtype: PIL.Image.Image
        """
        # eventlog.log_i("gr {}, r {}, a {}".format(green, red, amber ))

        w, h = im_rgba.wh
        with Image.frombuffer(
                'RGBA', (w, h + im_rgba.dh), im_rgba.buffer,
                'raw', 'RGBA', im_rgba.stride, 1) as pil_img:

            pil_img.readonly = 0

            im_draw_w, im_draw_h = im_rgba.wh

            im_draw = ImageDraw.Draw(pil_img)

            # TODO: render lamps, add lamps position in conf ?
            # green_time, red_time, amber_time = tl_service.get_times(tr_scene.im_ref.ts)
            # _render_tl(im_draw, green_time, red_time, amber_time)

            im_draw.font = Renderer._FONT

            map(lambda ra_item: self._render_ra_item(im_draw, im_draw_w, im_draw_h, self._ra_anno_args, ra_item),
                tr_scene.ra_items)

            map(lambda lr_item: self._render_lr_item(
                im_draw, im_draw_w, im_draw_h, self._lr_anno_args, lr_item,
                tr_scene.im_ref.width, tr_scene.im_ref.height),
                tr_scene.lr_items)

            map(lambda face_item: self._render_face_item(
                im_draw, im_draw_w, im_draw_h, tr_scene.im_ref.width, tr_scene.im_ref.height, face_item),
                tr_scene.face_items)

            map(lambda tr_item: self._render_tr_item(
                im_draw, im_draw_w, im_draw_h, tr_item,
                tr_scene.im_ref.width,
                tr_scene.im_ref.height
            ), tr_scene.tr_items)

            return pil_img

    def _render_face_item(self, draw, draw_w, draw_h, face_scene_w, face_scene_h, face_item):
        """
        :type draw: PIL.ImageDraw.ImageDraw
        :type draw_w: int
        :type face_scene_w: int
        :type face_scene_h: int
        :type draw_h: int
        :type face_item: FaceSceneItem
        """

        bbox = face_item.face_bbox

        ra_ltrb = (
            int(float(bbox.x0) / face_scene_w * draw_w),
            int(float(bbox.y0) / face_scene_h * draw_h),
            int(float(bbox.x1) / face_scene_w * draw_w),
            int(float(bbox.y1) / face_scene_h * draw_h)
            )

        draw.rectangle(ra_ltrb, outline=Renderer._COLOR_RED)

        # ltrb = _get_ra_board_box_ltrb((x, y), _RA_BOX_LANE)
        # rgb_image_draw.rectangle(ltrb, outline=Renderer._COLOR_RED)

    def _render_ra_item(self, draw, draw_w, draw_h, args, ra_item, ref_xy=None):
        """
        :type draw: PIL.ImageDraw.ImageDraw
        :type draw_w: int
        :type draw_h: int
        :type args: RaAnnoArgs
        :type ra_item: RaSceneItem
        :type ref_xy: (int, int) | None
        """
        if ref_xy:
            lr_ra_match = True
        else:
            ref_xy = ra_item.map_bbox_ref(draw_w, draw_h)
            lr_ra_match = False

        if args.show_pos(lr_ra_match):
            abs_x = int(ra_item.ra_rel_im_x * draw_w)
            abs_y = int(ra_item.ra_rel_im_y * draw_h)

            ra_ltrb = (
                abs_x - self._RA_ITEM_RADIUS,
                abs_y - self._RA_ITEM_RADIUS,
                abs_x + self._RA_ITEM_RADIUS,
                abs_y + self._RA_ITEM_RADIUS)

            draw.rectangle(ra_ltrb, fill=Renderer._COLOR_RED)

        # ltrb = _get_ra_board_box_ltrb((x, y), _RA_BOX_LANE)
        # rgb_image_draw.rectangle(ltrb, outline=Renderer._COLOR_RED)

        if args.show_len(lr_ra_match):
            bg_ltrb = _get_ra_board_box_ltrb(ref_xy, _RA_BOX_LENGTH)
            _render_text(draw, bg_ltrb, fg_text=str(ra_item.ra_len))

            if args.show_speed(lr_ra_match):
                bg_ltrb = _get_ra_board_box_ltrb(ref_xy, _RA_BOX_SPEED)
                _render_speed(draw, bg_ltrb, ra_item.ra_speed,
                              args.yellow_speed, args.red_speed)

            if args.show_dir(lr_ra_match):
                bg_ltrb = _get_ra_board_box_ltrb(ref_xy, _RA_BOX_DIR)
                _render_dir(draw, bg_ltrb, ra_item.ra_dir)

        else:
            if args.show_speed(lr_ra_match):
                bg_ltrb = _get_ra_board_box_ltrb(ref_xy, _RA_BOX_SPEED_WITHOUT_LENGTH)
                _render_speed(draw, bg_ltrb, ra_item.ra_speed,
                              args.yellow_speed, args.red_speed)

            if args.show_dir(lr_ra_match):
                bg_ltrb = _get_ra_board_box_ltrb(ref_xy, _RA_BOX_DIR_WITHOUT_LENGTH)
                _render_dir(draw, bg_ltrb, ra_item.ra_dir)

    def _render_lr_item(self, draw, draw_w, draw_h, args, lr_item,
                        lr_scene_w, lr_scene_h, ref_point_xy=None):
        """
        :type draw: PIL.ImageDraw.ImageDraw
        :type draw_w: int
        :type draw_h: int
        :type args: LrAnnoArgs
        :type lr_item: LrSceneItem
        :type lr_scene_w: int
        :type lr_scene_h: int
        """
        if args.show_trace:
            for x_index in range(0, len(lr_item.lr_trace) - 2, 2):
                # noinspection PyBroadException
                try:
                    image_x = lr_item.lr_trace[x_index] * draw_w / lr_scene_w
                    image_y = lr_item.lr_trace[x_index + 1] * draw_h / lr_scene_h
                    lr_rect_history_ltrb = (image_x - 1, image_y - 1, image_x + 1, image_y + 1)

                    color = self._color_store.get_lr_oid_color(lr_item.lr_oid)
                    draw.rectangle(lr_rect_history_ltrb, fill=color)
                except Exception:
                    pass

        if not ref_point_xy:
            ref_point_xy = lr_item.map_bbox_ref(lr_scene_w, lr_scene_h, draw_w, draw_h)

        if args.show_margin:
            width, height = lr_item.lr_quad.map_bbox_wh(lr_scene_w, lr_scene_h, draw_w, draw_h)

            lr_margin = lr_item.lr_margin
            new_upper_left_x = ref_point_xy[0] - (width * lr_margin.left)
            new_lower_right_x = ref_point_xy[0] + (width * lr_margin.right)

            new_upper_left_y = ref_point_xy[1] - (height * lr_margin.top)
            new_lower_right_y = ref_point_xy[1] + (height * lr_margin.bottom)
            new_plate_rect_ltrb = (new_upper_left_x, new_upper_left_y, new_lower_right_x, new_lower_right_y)
            draw.rectangle(new_plate_rect_ltrb, outline=Renderer._COLOR_YELLOW)

        if args.show_dir:
            bg_ltrb = _get_lr_board_box_ltrb(ref_point_xy, _LR_BOX_DIR)
            _render_dir(draw, bg_ltrb, lr_item.lr_dir)

        if args.show_lane:
            bg_ltrb = _get_lr_board_box_ltrb(ref_point_xy, _LR_BOX_LANE)
            lane_set = self._lane_detector.get_lane2(*lr_item.lr_quad.get_rel_bbox(lr_scene_w, lr_scene_h))
            lane = lane_set.pop() if len(lane_set) == 1 else _LANE_UNDEFINED
            lane_as_unicode = unicode(lane)
            if len(lane_as_unicode) > 1:
                lane_as_unicode = lane_as_unicode[0]
            _render_text(draw, bg_ltrb, fg_text=lane_as_unicode, bg_color=(206, 180, 141))

        if args.show_text:
            poly = lr_item.lr_quad.map(lr_scene_w, lr_scene_h, draw_w, draw_h,
                                       self._PLATE_MARGIN_Y, self._PLATE_MARGIN_X)
            draw.polygon(poly, outline=Renderer._COLOR_YELLOW)

            bg_ltrb = _get_lr_board_box_ltrb(ref_point_xy, _LR_BOX_PLATE_TEXT)
            _render_text(draw, bg_ltrb, lr_item.lr_text)

        if args.show_center:
            ref_point_xy2 = lr_item.map_bbox_ref2(lr_scene_w, lr_scene_h, draw_w, draw_h)

            x = ref_point_xy2[0]
            y = (ref_point_xy[1] + ref_point_xy2[1]) / 2

            bg_ltrb = _get_lr_board_box_ltrb(ref_point_xy2, _LR_BOX_COORD_BBOX)
            _render_text(draw, bg_ltrb, '{} {}'.format(x, y))

    def _render_tr_item(self, draw, draw_w, draw_h, tr_item, lr_scene_w, lr_scene_h):
        """
        :type draw: PIL.ImageDraw.ImageDraw
        :type draw_w: int
        :type draw_h: int
        :type tr_item: TrSceneItem
        :type lr_scene_w: int
        :type lr_scene_h: int
        """

        lr_item = tr_item.lr_item
        ref_point_x, ref_point_y = lr_item.map_bbox_ref(lr_scene_w, lr_scene_h, draw_w, draw_h)
        self._render_lr_item(draw, draw_w, draw_h, self._lr_anno_args,
                             lr_item, lr_scene_w, lr_scene_h,
                             (ref_point_x, ref_point_y + 2 * _SYM_MARGIN_H))

        ra_item = tr_item.ra_item
        ra_ref_xy = ref_point_x, ref_point_y + _SYM_H + 5 * _SYM_MARGIN_H
        self._render_ra_item(draw, draw_w, draw_h, self._ra_anno_args, ra_item, ra_ref_xy)
