from twisted.internet.defer import Deferred

from edge import eventlog
from edge_bbox.types import mk_plate_bbox2, QuadI, RectF
from edge.lane_detector import LaneDetector
from edge.fov import Fov

from edge.image.image import ImEngine, ImRgba, ImLoader, OverviewImEngine, ImBuffer
from edge.image.types import ImFooterVars, MemBlob
from edge.service.tz import ITzService
from edge.model.types import Dir, Location
from edge.service.about import IAboutService

from .types import DetectorArgs
from ..types import IBaseDetector, DetectorError
from ...model.types import (BaseEvent, Measurement, Waypoint, BestWaypoint,
                            Vehicle, LicensePlate, LicensePlateType, Position)
from ...model import models
from ...services.raas import IRaasService, RaasError
from ...services.trainer import ITrainerService
from ...services.plate_info import IPlateInfoService
from ...services.vs import IVsService, VsImageInfo, VsImageBufferError, VsError, VsImageBufferError2
from ...services.clock import IClockService


from .best import get_tr_item_dist


_UCODE_TO_ASCII_MAP = {
    ord(u'\u0410'): unichr(65),  # A
    ord(u'\u0412'): unichr(66),  # B
    ord(u'\u0421'): unichr(67),  # C
    ord(u'\u0415'): unichr(69),  # E
    ord(u'\u041d'): unichr(72),  # H
    ord(u'\u0406'): unichr(73),  # I
    ord(u'\u041a'): unichr(75),  # K
    ord(u'\u041c'): unichr(77),  # M
    ord(u'\u041e'): unichr(79),  # O
    ord(u'\u0420'): unichr(80),  # P
    ord(u'\u0422'): unichr(84),  # T
    ord(u'\u0425'): unichr(88)   # X
 }


def trans_plate_text(plate_text):
    """
    :type plate_text: str
    :rtype: str
    """
    return u''.join(map(lambda x: _UCODE_TO_ASCII_MAP.get(ord(x), x), plate_text))


def create(*args, **kwargs):
    """
    :rtype: IDetector
    """
    return _Detector(*args, **kwargs)


class _SceneVsRef(VsImageInfo):
    def __init__(self, tr_item, weight):
        """
        :type tr_item: models.TrTrackItem
        :type weight: int
        """
        VsImageInfo.__init__(self, tr_item.ts)
        self.tr_item = tr_item
        self.weight = weight


class _Detector(IBaseDetector):
    """
    :type _handler: None | (models.Track, BaseEvent) -> None
    """
    def __init__(
            self,
            tz_service,
            vs_service,
            raas_service,
            trainer_service,
            about_service,
            lane_detector,
            plate_info_service,
            fov_service,
            clock_service,
            args):
        """
        :type tz_service: ITzService
        :type vs_service: IVsService
        :type raas_service: IRaasService
        :type trainer_service: ITrainerService
        :type about_service: IAboutService
        :type lane_detector: LaneDetector
        :type plate_info_service: IPlateInfoService
        :type fov_service: Fov | None
        :type clock_service: IClockService
        :type args: DetectorArgs
        """
        self._tz = tz_service
        self._raas = raas_service
        self._trainer_service = trainer_service  # TODO
        self._about_service = about_service
        self._plate_info_service = plate_info_service
        self._clock_service = clock_service
        # create blobstore
        self._vs = vs_service

        self._rgba_scene_blob = ImRgba()
        self._wh = args.im_engine.scene.footer.height

        self._enable_overview = args.enable_overview
        self._overview_blob = ImBuffer()
        self._overview_engine = OverviewImEngine()

        self._lane_detector = lane_detector

        self._im_loader = ImLoader(args.im_engine.scene.footer.height)
        self._im_engine = ImEngine(args.im_engine)
        self._handler = None

        self._def_mapping = args.def_mapping
        self._fov_service = fov_service

        self._systest = args.systest

        IBaseDetector.__init__(self)

    def destroy(self):
        self._im_loader.destroy()

    def set_handler(self, event_handler):
        """
        :type event_handler: (models.Track, BaseEvent) -> None
        """
        if self._handler:
            raise NotImplementedError

        self._handler = event_handler

    def analyze(self, track):
        """
        :type track: models.Track
        """
        def mk_scene_ms_ref(tr_item):
            """
            :type tr_item: models.TrTrackItem
            :rtype: _SceneVsRef
            """
            tr_item = models.TrTrackItem(tr_item.ts, tr_item.lr_item, tr_item.ra_item, tr_item.le_item)
            lr_scene_w = lr_scene_h = None
            if track.lr_track:
                lr_scene_w = track.lr_track.best_image.width
                lr_scene_h = track.lr_track.best_image.width

            return _SceneVsRef(tr_item, get_tr_item_dist(tr_item, lr_scene_w, lr_scene_h))

        scene_ms_refs = map(mk_scene_ms_ref, track.tr_track)
        scene_ms_refs.sort(key=lambda ref: ref.weight, reverse=True)

        winner_img = None
        try:
            winner_img = self._vs.export_image_rgba(self._rgba_scene_blob, scene_ms_refs)
        except VsImageBufferError as ex:
            # print ex.min_w, ex.min_h
            self._rgba_scene_blob.reset(14, ex.min_w, ex.min_h, self._wh)
            winner_img = self._vs.export_image_rgba(self._rgba_scene_blob, scene_ms_refs)
        except VsError:
            pass

        if self._enable_overview:
            overview_img_info = None
            try:
                overview_img_info = self._vs.export_image_jpeg(self._overview_blob, scene_ms_refs, '1')
            except VsImageBufferError2 as ex:
                self._overview_blob.reset(14, ex.min_size)
                overview_img_info = self._vs.export_image_jpeg(self._overview_blob, scene_ms_refs, '1')
            except VsError:
                pass

            if overview_img_info is None:
                eventlog.log_e('fail: get overview image')

        if winner_img is None:
            if track.lr_track is None or track.lr_track.best_image.url is None:
                eventlog.log_d('fail: define scene-image (image-absent)')
                raise DetectorError

            lr_tr_scene_ref = track.lr_track.best_image
            try:
                self._im_loader.load_ref(lr_tr_scene_ref, self._rgba_scene_blob)
            except ValueError as x:
                eventlog.log_d('fail: load lpr image [%s]' % x.message)
                raise DetectorError

            except NotImplemented as x:
                eventlog.log_d('fail: unknown lpr image codec [%s]' % x.message)
                raise DetectorError

        self._analyze(track, winner_img.tr_item if winner_img else None)

    def _analyze(self, track, tr_item):
        """
        :type track: models.Track
        :type tr_item: models.TrTrackItem | None
        :rtype: Deferred
        """

        def get_lr_plate(raas_scene):
            """
            :type raas_scene: models.RaasScene
            :rtype: models.Plate
            :raise RaasError:
            """
            raas_plate = raas_scene.try_get_first_scene_item_plate()
            if raas_plate:
                return models.Plate(raas_plate.validity, raas_plate.text, raas_plate.template)

            eventlog.log_d('fail: define raas-result (scene is empty)')
            raise RaasError

        def use_lr_plate(failure):
            """
            :type failure: twisted.python.failure.Failure
            :rtype: models.Plate | None
            """
            # noinspection PyTypeChecker
            failure.trap(RaasError)

            if track.lr_track:
                return track.lr_track.plate

            return None

        def handle_fail(failure):
            """
            :type failure: twisted.python.failure.Failure
            """
            if failure.check([DetectorError]):
                eventlog.log_d('fail: process track')
            else:
                eventlog.log_d('fail: process track (%s)' % str(failure))

        if tr_item:
            rect = tr_item.get_plate_rect(*self._rgba_scene_blob.wh)
            quad = tr_item.get_plate_quad(*self._rgba_scene_blob.wh)
        else:
            best_img = track.lr_track.best_image
            rect = RectF(*track.lr_track.plate.quad.get_rel_bbox(best_img.width, best_img.height))
            quad = track.lr_track.plate.quad

        # Optimize plate recognition results
        call_chain = self._raas.analyze_scene2(rect, self._rgba_scene_blob)
        call_chain.addCallback(get_lr_plate)
        call_chain.addErrback(use_lr_plate)

        # Create an event.
        call_chain.addCallback(self._mk_event_args, track=track, tr_item=tr_item, quad=quad)
        call_chain.addCallback(lambda event_args: self._handler(track, event_args))
        call_chain.addErrback(handle_fail)
        return call_chain

    def _mk_event_args(self, plate, track, tr_item, quad):
        """
        :type plate: models.Plate | None
        :type track: models.Track
        :type tr_item: models.TrTrackItem | None
        :type quad: QuadI
        :rtype: BaseEvent
        """

        if tr_item:
            wh = self._rgba_scene_blob.wh
            wp_direction = track.tr_track.get_item_direction(tr_item)
        else:
            best_img = track.lr_track.best_image
            wh = best_img.width, best_img.height
            wp_direction = Dir(track.lr_track.direction)

        track_direction = track.get_direction()

        ts = self._rgba_scene_blob.ts

        def mk_position():
            """
            :rtype: Position | None
            """
            world_f, world_l = None, None
            if tr_item and tr_item.ra_item:
                world_f, world_l = tr_item.ra_item.raw_f, tr_item.ra_item.raw_l  # values is cm

            elif self._fov_service:
                plate_cx, plate_cy = quad.center
                world_f, world_l = self._fov_service.mk_world_point(plate_cx, plate_cy, world_t=0.5)
                # TODO: mirgate to new fov service to delete this workaround
                # m to cm transform
                world_f = int(world_f * 100)
                world_l = int(world_l * 100)
                #
                world_f = - world_f

            if None in [world_f, world_l]:
                return None
            else:
                return Position(world_f, world_l)

        def mk_vehicle():
            """
            :rtype: Vehicle
            """
            if plate:
                plate_type = LicensePlateType(plate.template, self._plate_info_service)
                # Ollie-Trans translator
                plate_text = trans_plate_text(plate.text)
                plates = [LicensePlate(plate.validity, plate_text, plate_type, quad)]
            else:
                plates = []

            spot_speed = track.ra_track.get_median_speed_kmph() if track.ra_track else None
            length = track.ra_track.length if track.ra_track else None
            position = mk_position()

            first_tr_item = None
            last_tr_item = None
            for index in range(0, len(track.tr_track), 1):
                i_tr_item = track.tr_track[index]  # type: models.TrTrackItem
                if i_tr_item.ra_item and i_tr_item.lr_item:
                    first_tr_item = i_tr_item
                    break

            for index in range(len(track.tr_track)-1, -1, -1):
                i_tr_item = track.tr_track[index]  # type: models.TrTrackItem
                if i_tr_item.ra_item and i_tr_item.lr_item:
                    last_tr_item = i_tr_item
                    break

            video_speed2 = None
            if None not in [first_tr_item, last_tr_item]:
                first_forward = first_tr_item.ra_item.raw_f / 100.0
                last_forward = last_tr_item.ra_item.raw_f / 100.0

                first_left = first_tr_item.ra_item.raw_l / 100.0
                last_left = last_tr_item.ra_item.raw_l / 100.0

                delta_f_m = last_forward - first_forward
                delta_l_m = last_left - first_left

                if delta_f_m >= 0:
                    delta_dist_m = pow(pow(delta_f_m, 2) + pow(delta_l_m, 2), 0.5)
                else:
                    delta_dist_m = - pow(pow(delta_f_m, 2) + pow(delta_l_m, 2), 0.5)

                delta_s_sec = (last_tr_item.ts - first_tr_item.ts) / 1e6

                video_speed2 = int(round(delta_dist_m / delta_s_sec * 3.6)) if delta_s_sec != 0 else None

            return Vehicle(plates, spot_speed, length, wp_direction, video_speed2, position,
                           track.lr_track.type_jnode if track.lr_track else None, systest=self._systest,
                           lpr_speed=track.lr_track.spot_speed if track.lr_track else None)

        best_origin = self._about_service.get_origin(ts)

        def mk_best_location():
            """
            :rtype: Location
            """
            if quad:
                lane_set = self._lane_detector.get_lane3(*quad.center)
                if len(lane_set) == 1:
                    return Location(best_origin.location.place, lane_set.pop())

            return Location(best_origin.location.place)

        vehicle = mk_vehicle()
        best_location = mk_best_location()

        scene_dh_data = self._rgba_scene_blob
        rect_f = RectF(*quad.get_rel_bbox(*wh))
        plate_bbox = mk_plate_bbox2(rect_f, scene_dh_data.w, scene_dh_data.h)
        control_ts = self._clock_service.get_control_ts(ts)

        def mk_best_images():
            """
            :rtype: list[MemBlob]
            """
            footer_vars = ImFooterVars(self._tz, self._about_service, ts, control_ts=control_ts, defs=self._def_mapping,
                                       location=best_location, geolocation=best_origin.geolocation,
                                       plate_text=vehicle.plate_text, spot_speed=vehicle.spot_speed,
                                       direction=wp_direction, radar_speed=vehicle.spot_speed,
                                       video_speed=vehicle.camera_speed, video_speed2=vehicle.video_speed2)

            blobs = self._im_engine.mk_transit_blobs(scene_dh_data, plate_bbox, footer_vars, track.lr_track, quad)
            return blobs

        best_images = mk_best_images()

        if self._enable_overview and self._overview_blob.ts > 14:
            ov_blob = self._overview_engine.mk_overview(self._overview_blob)
            best_images.append(ov_blob)

        # define #/measurement/0
        measurement = Measurement(track.uuid, direction=track_direction).add_waypoints([
            Waypoint(Waypoint.NAME_CATCH, track.tr_track.min_ts),
            Waypoint(Waypoint.NAME_LOST, track.tr_track.max_ts),
            BestWaypoint(ts, vehicle, control_ts=control_ts, direction=str(vehicle.direction),
                         location=best_location).add_images(best_images)
        ])

        base_event = BaseEvent([measurement], track.lr_track.lr_alarms_jnode if track.lr_track else None)

        def get_blob():
            return self._im_engine.mk_trainer_scene_blob(scene_dh_data)

        self._trainer_service.collect(base_event, get_blob, rect_f, track)

        return base_event
