from edge.image.types import ImEngineArgs
from edge.service.tz import ITzService
from edge.model.types import Dir, Origin

from ...model.types import Identity, Evidence, SceneArgs, BestViolationWaypoint
from ...detectors.com.types import ScheduleArgs, VideoArgs


class LimitsArgs:
    def __init__(self, vehicle_types):
        """
        :type vehicle_types: list of str
        """
        self.vehicle_types = vehicle_types


class RuleArgs:
    def __init__(self, name, enabled, schedule, ticket, limits):
        """
        :type name: str
        :type enabled: bool
        :type schedule: ScheduleArgs | None
        :type ticket: Ticket
        :type limits: LimitsArgs
        """
        self.name = name
        self.enabled = enabled
        self.schedule = schedule
        self.ticket = ticket
        self.limits = limits


class DetectorArgs:
    def __init__(self, identity, name, store, rules, def_footer_mapping, scene, im_engine, video_args):
        """
        :type identity: Identity
        :type name: str
        :type store: str
        :type rules: list of RuleArgs
        :type def_footer_mapping: dict
        :type scene: SceneArgs
        :type im_engine: ImEngineArgs
        :type video_args: VideoArgs | None
        """
        self.identity = identity
        self.name = name
        self.store = store
        self.rules = rules
        self.def_footer_mapping = def_footer_mapping
        self.scene = scene
        self.im_engine = im_engine
        self.video = video_args


class BusLaneEvidence(Evidence):
    def __init__(self, origin, ts, duration, args, detector_name):
        """
        :type origin: Origin
        :type ts: long
        :type duration: int
        :type args: RuleArgs
        :type detector_name: str
        """
        Evidence.__init__(self, Evidence.NAME_BUS_LANE, origin, ts)
        self.detector_name = detector_name
        self.args = args
        self.duration = duration

    def serialize(self, tz_service):
        """
        :type tz_service: ITzService
        :rtype: dict
        """
        jnode = Evidence.serialize(self, tz_service)

        violation_waypoints_jnode = list()
        best_waypoints_jnode = None
        for waypoint in self._waypoints:
            if waypoint.name == waypoint.NAME_VIOLATION:
                violation_waypoints_jnode.append(waypoint.serialize(tz_service))
            elif waypoint.name == waypoint.NAME_VIOLATION_BEST:
                best_waypoints_jnode = waypoint.serialize(tz_service)

        jnode['evidence'] = {
            'detector': {
                'name': self.detector_name,
                'rule': {
                    'name': self.args.name,
                    'schedule': self.args.schedule.serialize(),
                    'ticket': {
                        'name': self.args.ticket.name
                    },
                    'limits': {
                        'types': self.args.limits.vehicle_types
                    }
                }
            },
            'measurements': {
                'duration': self.duration,
                'waypoints': {
                    'violation': violation_waypoints_jnode
                }
            }
        }

        if best_waypoints_jnode is not None:
            jnode['evidence']['measurements']['waypoints']['best'] = best_waypoints_jnode

        if len(self._videos):
            videos_jnode = map(lambda x: x.serialize(tz_service), self._videos)
            jnode['evidence']['measurements']['videos'] = videos_jnode

        return jnode

    @property
    def best_waypoint(self):
        """
        :rtype: BestViolationWaypoint | None
        """
        bw = filter(lambda x: x.name == BestViolationWaypoint.NAME_VIOLATION_BEST, self.waypoints)
        return bw[0] if len(bw) else None
