from edge import eventlog
from edge.image.jnode.types import JsonImEngineArgs
from edge.types import JnodeError

from ..types import DetectorArgs, RuleArgs, LimitsArgs
from ...types import IDetector
from ....model.types import Identity
from ...com.jnode.types import JsonBaseRuleArgs, JsonVideoArgs


class JsonLimitsArgs(LimitsArgs):
    def __init__(self, *args, **kwargs):
        LimitsArgs.__init__(self, *args, **kwargs)

    @staticmethod
    def deserialize(limits_jnode):
        """
        :type limits_jnode: dict
        :rtype: JsonLimitsArgs
        :raise JnodeError:
        """
        # deserialize #/speed-limit
        speed_limit = limits_jnode.get('speed-limit')
        if not (int == type(speed_limit) and 0 <= speed_limit <= 250):
            eventlog.log_e_param('#/speed-limit')
            raise JnodeError

        # deserialize #/max-overspeed
        max_overspeed = limits_jnode.get('max-overspeed')
        if not (int == type(speed_limit) and 0 <= max_overspeed <= 250):
            eventlog.log_e_param('#/max-overspeed')
            raise JnodeError

        # deserialize #/lanes
        lanes_jnode = limits_jnode.get('lanes', [])

        if type(lanes_jnode) != list:
            eventlog.log_e_param('#/lanes')
            raise JnodeError

        for lane in lanes_jnode:
            if not type(lane) in [str, unicode]:
                eventlog.log_e_param('#/lanes/%s' % lane)
                raise JnodeError

        return JsonLimitsArgs(lanes_jnode, speed_limit, max_overspeed)


class JsonRuleArgs(RuleArgs):
    def __init__(self, *args, **kwargs):
        RuleArgs.__init__(self, *args, **kwargs)

    @staticmethod
    def deserialize(rule_jnode):
        """
        :type rule_jnode: dict
        :rtype: JsonRuleArgs
        :raise JnodeError:
        """
        if type(rule_jnode) != dict:
            eventlog.log_e_param('#/')
            raise JnodeError

        # deserialize #/name
        # deserialize #/enabled
        # deserialize #/schedule
        # deserialize #/ticket
        rule = JsonBaseRuleArgs.deserialize(rule_jnode)

        # deserialize #/limits
        limits_jnode = rule_jnode.get('limits')
        if type(rule_jnode) != dict:
            eventlog.log_e_param('#/limits')
            raise JnodeError

        limits = JsonLimitsArgs.deserialize(limits_jnode)

        return JsonRuleArgs(
            rule.name, rule.enabled, rule.schedule, rule.ticket,
            limits)


class JsonDetectorArgs(DetectorArgs):
    def __init__(self, *args, **kwargs):
        DetectorArgs.__init__(self, *args, **kwargs)

    @staticmethod
    def deserialize(jnode, index, def_footer_mapping):
        """
        :type jnode: dict
        :type index: int
        :type def_footer_mapping: dict
        :rtype: JsonDetectorArgs
        :raise JnodeError:
        """
        # deserialize #/name
        name = jnode.get('name')
        if not type(name) in [str, unicode] or len(name) > 32:
            eventlog.log_e_param('#/name')
            raise JnodeError

        # deserialize #/store
        store = jnode.get('store')
        if not (type(store) in [str, unicode] and 1 <= len(store) <= 2):
            eventlog.log_e_param('#/store')
            raise JnodeError

        # deserialize #/rules
        rules_jnode = jnode.get('rules')
        if type(rules_jnode) != list:
            eventlog.log_e_param('#/rules')
            raise JnodeError

        rules = map(JsonRuleArgs.deserialize, rules_jnode)
        identity = Identity(IDetector.TYPE_SPOT_SPEED, index)

        # deserialize #/images
        images_jnode = jnode.get('images')
        if type(images_jnode) != dict:
            eventlog.log_e_param('#/images')
            raise JnodeError

        im_engine = JsonImEngineArgs.deserialize(images_jnode['violation'])

        # deserialize #/videos/violation
        video_jnode = jnode.get('videos', {}).get('violation')
        if video_jnode is not None:
            if video_jnode.get('enabled', False):
                video_args = JsonVideoArgs.deserialize(video_jnode)
            else:
                video_args = None
        else:
            video_args = None

        # deserialize #/h-over-ground
        plate_over_gound = jnode.get('plate-over-ground', None)
        log = jnode.get('log-cam-speed', False)

        return JsonDetectorArgs(identity, name, store, rules, im_engine, def_footer_mapping, video_args,
                                plate_over_gound, log)

