import ctypes

from glob import glob
from edge import eventlog
from edge.types import BufferSizeError
import ujson
from .types import SignServiceArgs
from .stats_provider import IStatsProvider


class SignError(RuntimeError):
    pass


class ISignService(IStatsProvider):

    def destroy(self):
        pass

    @property
    def cert(self):
        """
        :rtype: str
        """
        raise NotImplementedError

    @property
    def hash_type(self):
        """
        :rtype: str
        """
        raise NotImplementedError

    def hash_blob(self, blob_buffer, hash_buffer):
        """
        :type blob_buffer: str
        :type hash_buffer: str

        :rtype: int
        :raise SignError:
        """
        raise NotImplementedError

    def sign_blob(self, blob_buffer, sign_handler):
        """
        :type blob_buffer: str
        :type sign_handler: Any

        :rtype: int
        :raise SignError:
        """
        raise NotImplementedError

    def collect_stats(self):
        raise NotImplementedError


class _StubService(ISignService):
    @property
    def cert(self):
        return ''

    @property
    def hash_type(self):
        return ''

    def hash_blob(self, blob_buffer, hash_buffer):
        return 0

    def sign_blob(self, blob_buffer, sign_handler):
        sign_handler(None, 0)
        return 0

    def collect_stats(self):
        return []


class _CryptoApi:
    LOG_N = 0
    LOG_E = 1
    LOG_W = 2
    LOG_I = 3
    LOG_D = 4
    LOG_T = 5

    LOG_FLAGS_NONE = 0
    LOG_FLAGS_STDERR = 1

    RC_DONE = 0
    RC_FAIL = -1
    RC_DENY = -2

    _lib = None
    _libpath = (glob('/opt/edge/lib/libsigniit.so') + [None])[0]

    if _libpath:
        _lib = ctypes.CDLL(_libpath)

        # void CryptoLogOpen(const char *aName, int aMaxLevel, int aFlags);
        logopen = _lib.CryptoLogOpen
        logopen.restype = ctypes.c_void_p
        logopen.argtypes = (ctypes.c_char_p, ctypes.c_int, ctypes.c_int)

        # int
        # CryptoProviderOpen(const char *aConfPath)
        open = _lib.CryptoProviderOpen
        open.restype = ctypes.c_int
        open.argtypes = (ctypes.c_char_p,)

        # int
        # CryptoProviderClose()
        close = _lib.CryptoProviderClose
        close.argtypes = ()

        # int
        # CryptoProviderStart()
        start = _lib.CryptoProviderStart
        start.argtypes = ()

        # int
        # CryptoProviderStop()
        stop = _lib.CryptoProviderStop
        stop.argtypes = ()

        # int
        # CryptoProviderGetCrt(uint8_t *aCrtData, uint32_t *aCrtSize);
        getcrt = _lib.CryptoProviderGetCrt
        getcrt.restype = ctypes.c_int
        getcrt.argtypes = (ctypes.POINTER(ctypes.c_uint8), ctypes.POINTER(ctypes.c_uint32))

        # typedef void (*sign_handler_fn)(void *aHandlerArg, int aStat, const uint8_t *aSignData, uint32_t aSignSize);
        SignHandler = ctypes.CFUNCTYPE(
            None,
            ctypes.c_void_p, ctypes.c_int,
            ctypes.POINTER(ctypes.c_uint8), ctypes.c_uint32)

        # typedef void (*crypto_stats_handler_fn)(void *aHandlerArg, const char *aJson);
        StatsHandler = ctypes.CFUNCTYPE(
            None,
            ctypes.c_void_p, ctypes.POINTER(ctypes.c_char)
        )

        # void
        # CryptoProviderCollectStats(crypto_stats_handler_fn aHandler, void *aHandlerArg);
        collect_stats = _lib.CryptoProviderCollectStats
        collect_stats.restype = ctypes.c_void_p
        collect_stats.argtypes = (StatsHandler, ctypes.c_void_p)

        # int
        # CryptoProviderSign(const uint8_t *aBlobData, uint32_t aBlobSize,
        #     sign_handler_fn aHandler, void *aHandlerArg)
        sign = _lib.CryptoProviderSign
        sign.restype = ctypes.c_int
        sign.argtypes = (ctypes.POINTER(ctypes.c_uint8), ctypes.c_uint32, SignHandler, ctypes.c_void_p)

        # int
        # CryptoProviderHash (const uint8_t *aBlobData, uint32_t aBlobSize,
        #     uint8_t *aHashData, uint32_t *aHashSize);
        hash = _lib.CryptoProviderHash
        hash.restype = ctypes.c_int
        hash.argtypes = (ctypes.POINTER(ctypes.c_uint8), ctypes.c_uint32,
                         ctypes.POINTER(ctypes.c_uint8), ctypes.POINTER(ctypes.c_uint32))

    hash_type = 'gost34311'

    def __init__(self):
        pass

    @classmethod
    def log_level_name_to_code(cls, name):
        name_to_code = {
            'none': cls.LOG_N,
            'error': cls.LOG_E,
            'warn': cls.LOG_W,
            'info': cls.LOG_I,
            'debug': cls.LOG_D,
            'trace': cls.LOG_T
        }

        return name_to_code.get(name, cls.LOG_I)

    @classmethod
    def test_available(cls):
        """
        :rtype: bool
        """
        return cls._lib is not None


class _SignWork:
    """
    :type _service: _SignService | None
    """
    _ref_to_self_map = dict()

    def __init__(self, service, blob_buffer, handler):
        """
        :type service: _SignService
        :type handler: (FileBlob | None, int) -> None
        """
        self._ref = id(self)
        self._handler = handler
        self._service = service

        #
        def sign_handler_py(work_ref, rec_stat, sign_data, sign_size):
            """
            :type work_ref: int
            :type rec_stat: int
            :type sign_data: ctypes.POINTER(ctypes.c_uint8)
            :type sign_size: int
            :rtype: None
            """
            eventlog.log_d('exec: fini sign-work (ref:%s)' % work_ref)
            work = _SignWork._ref_to_self_map.pop(work_ref, None)  # type: _SignWork
            if work is None:
                eventlog.log_w('fail: fini sign-work (ref:%s)' % work_ref)
                return

            else:
                eventlog.log_d('done: fini sign-work (ref:%s)' % work_ref)

            #
            if rec_stat != _CryptoApi.RC_DONE:
                eventlog.log_e('fail: CryptoProviderSign (stat:%s)' % rec_stat)
                work._run_async_handler(None, None)

            else:
                eventlog.log_d('done: CryptoProviderSign (stat:%s)' % rec_stat)
                work._run_async_handler(sign_data, sign_size)

        self._sign_handler_c = _CryptoApi.SignHandler(sign_handler_py)

        #
        self._run_args = (blob_buffer, self._sign_handler_c, self._ref)

    def __del__(self):
        if _SignWork._ref_to_self_map.pop(self._ref, None):
            eventlog.log_i('done: fini sign-work (ref:%s)' % self._ref)

    def run_async(self):
        """
        :rtype: None
        :raise SignError:
        """
        _SignWork._ref_to_self_map[self._ref] = self
        eventlog.log_d('exec: init sign-work (ref:%s)' % self._ref)

        # noinspection PyBroadException
        try:
            self._service.sign_buffer(*self._run_args)
            eventlog.log_d('done: init sign-work (ref:%s)' % self._ref)
            return
        except SignError:
            eventlog.log_e('fail: init sign-work (ref:%s, session error' % self._ref)
        except Exception:
            eventlog.log_x()

        del _SignWork._ref_to_self_map[self._ref]
        eventlog.log_d('done: fini sign-work (ref:%s)' % self._ref)

        raise SignError

    def _run_async_handler(self, sign_data, sign_size):
        """
        :type sign_data: FileBlob | None
        :rtype: None
        """
        eventlog.log_d('exec: ref:{}, blob:{}'.format(
            self._ref, sign_data if sign_data else 'none'))

        # noinspection PyBroadException
        try:
            self._handler(sign_data, sign_size)
            eventlog.log_d('done: ref:{}'.format(self._ref))
        except Exception:
            eventlog.log_x()


class _SignService(ISignService):
    def __init__(self, crypto_api):
        """
        :type crypto_api: _CryptoApi
        """
        self._api = crypto_api
        self._cert = None

    def destroy(self):
        self._api.stop()
        self._api.close()

    @property
    def cert(self):
        """
        :rtype: str
        """
        if self._cert is None:
            buffer_data = ctypes.create_string_buffer('', size=1)
            for _ in range(2):
                try:
                    buffer_data2 = ctypes.cast(buffer_data, ctypes.POINTER(ctypes.c_uint8))

                    buffer_size = len(buffer_data)
                    buffer_size_used = ctypes.c_uint32(buffer_size)
                    buffer_size_used2 = ctypes.pointer(buffer_size_used)

                    rc = self._api.getcrt(buffer_data2, buffer_size_used2)
                    if rc == self._api.RC_DENY:
                        eventlog.log_e('fail: GetCrt <RC_DENY>')
                        raise SignError

                    elif rc != self._api.RC_DONE:
                        eventlog.log_e('fail: GetCrt <{}>'.format(rc))
                        raise SignError

                    buffer_used = buffer_size_used.value
                    if buffer_used > buffer_size:
                        eventlog.log_e('fail: crt-buffer too small (min-size: {})'.format(buffer_used))
                        raise BufferSizeError(buffer_used)

                    self._cert = buffer_data.raw
                    break

                except BufferSizeError as x:
                    buffer_data = ctypes.create_string_buffer('', size=x.min_size)

        return self._cert

    @property
    def hash_type(self):
        return self._api.hash_type

    def hash_blob(self, blob_buffer, hash_buffer):
        """
        :type blob_buffer: str
        :param hash_buffer: ctypes.create_string_buffer
        :type hash_buffer: str

        :rtype: int
        :raise SignError:
        """

        blob_data = ctypes.cast(blob_buffer, ctypes.POINTER(ctypes.c_uint8))
        blob_size = ctypes.c_uint32(len(blob_buffer))

        buffer_data = ctypes.cast(hash_buffer, ctypes.POINTER(ctypes.c_uint8))

        buffer_size = len(hash_buffer)
        buffer_size_used = ctypes.c_uint32(buffer_size)
        buffer_size_used2 = ctypes.pointer(buffer_size_used)

        rc = self._api.hash(blob_data, blob_size, buffer_data, buffer_size_used2)
        if rc == self._api.RC_DENY:
            eventlog.log_e('fail: Hash <RC_DENY>')
            raise SignError

        elif rc != self._api.RC_DONE:
            eventlog.log_e('fail: Hash (%s)' % rc)
            raise SignError

        buffer_used = buffer_size_used.value
        if buffer_used > buffer_size:
            eventlog.log_e('fail: hash-buffer too small (min-size: {})'.format(buffer_used))
            raise BufferSizeError(buffer_used)

        return buffer_used

    def sign_blob(self, blob_buffer, sign_handler):
        try:
            sign_work = _SignWork(self, blob_buffer, sign_handler)
            sign_work.run_async()
        except SignError:
            eventlog.log_e('fail: session error')
            raise
        except Exception:
            eventlog.log_x()
            raise SignError

    def sign_buffer(self, blob_buffer, cb, cb_arg):
        """
        :type blob_buffer: str
        :param cb: _Api.SignHandler
        :type cb: Any
        :type cb_arg: int

        :rtype: None
        :raise SignError:
        """
        blob_data = ctypes.cast(blob_buffer, ctypes.POINTER(ctypes.c_uint8))
        blob_size = ctypes.c_uint32(len(blob_buffer))

        rc = self._api.sign(blob_data, blob_size, cb, cb_arg)
        if rc == self._api.RC_DENY:
            eventlog.log_e('fail: Sign <RC_DENY>')
            raise SignError

        elif rc != self._api.RC_DONE:
            eventlog.log_e('fail: Sign <answer:{}>'.format(rc))
            raise SignError

    def collect_stats(self):
        test_list = []

        def stat_handler_py(_, stat_data):
            test = ctypes.string_at(stat_data)
            test_list.append(test)

        _sign_handler_c = _CryptoApi.StatsHandler(stat_handler_py)

        self._api.collect_stats(_sign_handler_c, 14)
        reply = ujson.loads(test_list[0])
        return reply


def init_service(conf):
    """
    :type conf: SignServiceArgs | None
    :rtype: ISignService
    """
    if conf is None:
        eventlog.log_i('using _StubService')
        return _StubService()

    elif not _CryptoApi.test_available():
        eventlog.log_i('using _StubService')
        return _StubService()

    else:
        eventlog.log_i('using libsign-iit')

    log_level = _CryptoApi.log_level_name_to_code(conf.log_level)
    _CryptoApi.logopen('speedcontrol.iit', log_level, _CryptoApi.LOG_FLAGS_NONE)

    rc = _CryptoApi.open(conf.iit_conf)
    _CryptoApi.start()
    if rc == _CryptoApi.RC_DENY:
        eventlog.log_e('fail: Open <RC_DENY>')
        raise SignError

    elif rc != _CryptoApi.RC_DONE:
        eventlog.log_e('fail: Open <{}>'.format(rc))
        raise SignError

    return _SignService(_CryptoApi)


def fini_service(service):
    """
    :type service: ISignService | _SignService
    """
    service.destroy()
